import time
from deeplearning.utils.tqdm import trange
from PIL import Image
import numpy as np
import os
import joblib

def validate(current_model, val_fn, epoch, num_batches_val, val_its, debug_folder="", tta=False):
    epoch_val_err = 0
    epoch_val_acc = 0
    epoch_conf_mat =  None
    debug_img_cnt = 0    
    for i in trange(num_batches_val):
        inputs, targets, locations, image_names = val_its[0].getBatch()
        if len(val_its) > 1:
            for val_it in val_its[1:]:
                new_inputs, new_targets, new_locations, new_image_names = val_it.getBatch()                    
                inputs = np.concatenate((inputs, new_inputs))
                targets = np.concatenate((targets, new_targets))  
                if locations is not None:
                    locations = np.concatenate((locations, new_locations))
                if image_names is not None:
                    image_names = np.concatenate((image_names, new_image_names))
        if len(inputs.shape) > 4:
            inputs = np.split((inputs/255.).astype('float32'), inputs.shape[1], axis=1)
            inputs = [np.squeeze(inp)[None].transpose(1,0,2,3) for inp in inputs]
            arguments = inputs + [targets]
        else:
            arguments = [(inputs/255.).astype('float32'), targets]          
        if tta:
            batch_val_err = 0
            batch_val_preds = None
            for k in xrange(4):
                arguments[:-1] = [np.rot90(x.transpose(2,3,0,1), k).transpose(2,3,0,1) for x in arguments[:-1]]
                tta_val_err, _, tta_val_preds = val_fn(*arguments)
                batch_val_err += (tta_val_err / 4.)
                if batch_val_preds is None:
                    batch_val_preds = (tta_val_preds / 4.)
                else:
                    batch_val_preds += (tta_val_preds / 4.)
                batch_val_acc = (np.argmax(targets, axis=1) == np.argmax(tta_val_preds, axis=1)).sum() / float(targets.shape[0])
                
        else:
            batch_val_err, batch_val_acc, batch_val_preds= val_fn(*arguments)
        epoch_val_err += batch_val_err
        epoch_val_acc += batch_val_acc
        if epoch_conf_mat is None:
            epoch_conf_mat = np.zeros((batch_val_preds.shape[1], batch_val_preds.shape[1]), dtype="float32")
        np.add.at(epoch_conf_mat, [np.argmax(targets, axis=1), np.argmax(batch_val_preds, axis=1)], 1)
        if debug_folder:
            labels = arguments[-1]
            for i in range(arguments[0].shape[0]):
                img_label = np.argmax(labels[i])
                img_name = os.path.split(image_names[i])[1]
                img_loc = locations[i]
                img_loss = 1. - batch_val_preds[i, img_label]
                if img_loss > 0.:
                    img_loss_str = str(img_loss)[2:6]
                    for inp_nr, inp in enumerate(arguments[:-1]):
                        img_name = img_name + "_x" + str(img_loc[0]) + "_y" + str(img_loc[1]) + "_epoch" + str(epoch) + "_nr" + str(debug_img_cnt) + "_class" + str(img_label) + "_loss" + img_loss_str + "_input" + str(inp_nr) + ".png"
                        conv_arr = (inp[i].transpose(1,2,0) * 255).astype('ubyte')
                        conv_img = Image.fromarray(conv_arr)
                        conv_img.save(os.path.join(debug_folder, img_name))
                debug_img_cnt += 1
    norm_epoch_conv_mat = epoch_conf_mat / np.sum(epoch_conf_mat, axis=1)[:, None]    
    print("  validation loss:\t\t{:.6f}".format(epoch_val_err / num_batches_val))
    print("  validation accuracy:\t\t{:.2f} %".format(epoch_val_acc / num_batches_val * 100))
    if tta:
        print("  with test-time augmentation")
    print("  validation confusion matrix:")
    print norm_epoch_conv_mat
    return epoch_val_err, epoch_val_acc

def train(current_model, train_fn, val_fn, num_epochs, num_batches_tra,
          num_batches_val, train_its, val_its, booster=None,
          model_save_folder=None, prev_epoch_dict=None, debug_folder=None,
          learning_rate=None, learning_rate_decay=None):
    if debug_folder:
        if os.path.isdir(debug_folder):
            print "In debug mode, writing to folder debug_folder"
    if prev_epoch_dict:
        epochs = prev_epoch_dict['epochs']
        total_batches = prev_epoch_dict['trained_batches']
        train_loss = prev_epoch_dict['train_loss']
        if prev_epoch_dict.has_key("train_reg_loss"):
            train_reg_loss = prev_epoch_dict["train_reg_loss"]
        else:
            train_reg_loss = [0] * len(prev_epoch_dict['train_acc'])
        train_acc = prev_epoch_dict['train_acc']
        val_loss = prev_epoch_dict['val_loss']
        val_acc = prev_epoch_dict['val_acc']
        print "Previous epoch was " + str(epochs[-1])
        print "Training loss at previous iteration: " + str(train_loss[-1])
        print "Validation loss at previous iteration: " + str(val_loss[-1])    
        print "Training accuracy at previous iteration: " + str(train_acc[-1])
        print "Validation accuracy at previous iteration: " + str(val_acc[-1])
        start_epoch= epochs[-1]+1
    else:
        train_loss = []
        train_reg_loss = []
        train_acc = []
        val_loss = []
        val_acc = []
        epochs = [0]
        total_batches = [0]
        start_epoch = 0
    for epoch in range(start_epoch, num_epochs):
        if epoch != 0 and epoch % 10 == 0 and learning_rate and learning_rate_decay:
            learning_rate = learning_rate * learning_rate_decay
            print "Learning rate decayed to: " + str(learning_rate)
        # In each epoch, we do a full pass over the training data:
        epoch_train_error = 0
        epoch_train_acc = 0
        epoch_train_reg_error = 0.
        train_batches = 0
        start_time = time.time()
        avg_btch_time = 0.
        avg_class_time = 0.
        for i in trange(num_batches_tra):
            btch_start_time = time.time()
            inputs, targets, locations, image_names = train_its[0].getBatch()
            if len(train_its) > 1:
                for train_it in train_its[1:]:
                    new_inputs, new_targets, new_locations, new_image_names = train_it.getBatch()                    
                    inputs = np.concatenate((inputs, new_inputs))
                    targets = np.concatenate((targets, new_targets))
                    if locations is not None:
                        locations = np.concatenate((locations, new_locations))
                    if image_names is not None:
                        image_names = np.concatenate((image_names, new_image_names))                    
            avg_btch_time += time.time() - btch_start_time
            inputs = (inputs/255.).astype('float32')
            if booster:
              inputs, targets = booster.preprocess_batch((inputs, targets))            
            if len(inputs.shape) > 4:
                inputs_split = np.split(inputs, inputs.shape[1], axis=1)
                inputs_split = [np.squeeze(inp)[None].transpose(1,0,2,3) for inp in inputs_split]
                arguments = inputs_split + [targets]
            else:
                arguments = [inputs, targets]
            class_start_time = time.time()
            train_outputs = train_fn(*arguments)
            batch_train_err, batch_train_acc, batch_preds = train_outputs[:3]
            if len(train_outputs) > 3:
                epoch_train_reg_error += train_outputs[3]                
            avg_class_time += time.time() - class_start_time
            if booster:
              booster.postprocess_batch((inputs, targets), batch_preds);
            epoch_train_error += batch_train_err
            epoch_train_acc += batch_train_acc
            train_batches += 1
        total_batches.append(train_batches + total_batches[-1])
        print("Epoch {} of {} took {:.3f}s".format(
            epoch + 1, num_epochs, time.time() - start_time))
        print("Average time to get a batch took {:.3f}s".format(
            avg_btch_time / train_batches))
        print("Average time to classsify a batch took {:.3f}s".format(
            avg_class_time / train_batches))
        print("  training loss:\t\t{:.6f}".format(epoch_train_error / train_batches))
        if epoch_train_reg_error > 0:
            print("  L2 loss:\t\t\t{:.6f}".format(epoch_train_reg_error / train_batches))
        print("  training accuracy:\t\t{:.2f} %".format(
            epoch_train_acc / train_batches * 100))

        # And a full pass over the validation data:
        epoch_val_err, epoch_val_acc = validate(current_model, val_fn, epoch, num_batches_val, val_its, debug_folder="")
        train_loss.append(epoch_train_error / train_batches)
        train_reg_loss.append(epoch_train_reg_error / train_batches)
        train_acc.append(epoch_train_acc / train_batches)
        val_loss.append(epoch_val_err / num_batches_val)
        val_acc.append(epoch_val_acc / num_batches_val)
        epochs.append(epoch)
        if model_save_folder:
            epoch_name = "Model_At_Epoch_" + ("{:0"+str(len(str(num_epochs)))+"d}").format(epoch) + ".pkl"
            joblib.dump({"model" : current_model, "epochs" : epochs, "trained_batches":total_batches, "train_loss":train_loss, "train_reg_loss" : train_reg_loss, "train_acc":train_acc, "val_loss":val_loss, "val_acc":val_acc}, os.path.join(model_save_folder, epoch_name), 9)
    return train_loss, train_acc, val_loss, val_acc